/******************************************************************
BigAtom Utilities by Shadowdarke (shadowdarke@hotmail.com) April 2006
bigatom_LocForPart() and bigatom_MapBoundaryCheck() procs by Deadron

These procs are nonessential expansions for the basic BigAtom library.
Including the library will not automatically include these procs.

You may include them by adding the following line to your project:
	#include <BigAtomUtilies.dm>

New atom/movable/procs:
	bigatom_Cull(flags)
		Removes unneeded atoms from the bigatom_parts list. This function only works with
			big atoms in overlay mode.
		ARGS:
			flags - Tells the proc how to determine what atoms are important. Accepts
				the following values:
					BIGATOM_KEEP_EDGES		(1) - keep parts on the outside edge of the big atom
					BIGATOM_KEEP_CORNERS	(2) - keep parts in the corners of the big atom
					BIGATOM_KEEP_DENSE		(4) - keep parts that are dense
					BIGATOM_KEEP_OPAQUE		(8) - Keep parts that are opaque
				The flags may be combined by the OR (|) operator to keep parts based on multiple
					criteria. For example, bigatom_Cull(BIGATOM_KEEP_DENSE | BIGATOM_KEEP_EDGES)
					will preserve any parts that are dense or on the edge of the big atom
				DEFAULT: flags = BIGATOM_KEEP_CORNERS | BIGATOM_KEEP_DENSE | BIGATOM_KEEP_OPAQUE

	bigatom_LocForPart(atom/movable/part, turf/Location)
		For the given destination, where should this part be placed relative to bigatom_src?

	bigatom_MapBoundaryCheck(turf/Location)
		Returns 1 if the big atom will be entirely on the map at this location, zero if not.

******************************************************************/

#define BIGATOM_KEEP_EDGES		1
#define BIGATOM_KEEP_CORNERS	2
#define BIGATOM_KEEP_DENSE		4
#define BIGATOM_KEEP_OPAQUE		8

atom/movable/proc
	bigatom_Cull(flags = BIGATOM_KEEP_CORNERS | BIGATOM_KEEP_DENSE | BIGATOM_KEEP_OPAQUE)
		// remove unneeded atoms from the parts list
		if(bigatom_src != src)
			if(bigatom_src) bigatom_src.bigatom_Cull(flags)
			return
		if(bigatom_mode & BIGATOM_NOOVERLAYS)
			// can only cull in overlays mode
			return
		for(var/atom/movable/M in bigatom_parts)
			if(density && (flags & BIGATOM_KEEP_DENSE)) continue
			if(opacity && (flags & BIGATOM_KEEP_OPAQUE)) continue
			if(flags & BIGATOM_KEEP_EDGES)
				var/pos_x = M.bigatom_x - bigatom_x + 1
				var/pos_y = M.bigatom_y - bigatom_y + 1
				if(!pos_x || (pos_x == bigatom_width) || \
					!pos_y || (pos_y == bigatom_height)) continue
			else if(flags & BIGATOM_KEEP_CORNERS)
				var/pos_x = M.bigatom_x - bigatom_x + 1
				var/pos_y = M.bigatom_y - bigatom_y + 1
				if((!pos_x || (pos_x == bigatom_width)) && \
					(!pos_y || (pos_y == bigatom_height))) continue
			del(M)

	bigatom_LocForPart(atom/movable/part, turf/Location)
		/* For the given destination, where should this part be placed relative to
			bigatom_src? */
		var/column_offset = Location.x + part.bigatom_x
		var/row_offset = Location.y + part.bigatom_y

		var/turf/destturf = locate(column_offset, row_offset, Location.z)
		if (!destturf || !destturf.Enter(part))
			return 0
		return destturf

	bigatom_MapBoundaryCheck(turf/Location)
		// Returns 1 if the object will be entirely on the map at this location, zero if not.
		var/top_x = Location.x + (bigatom_width - 1)
		var/top_y = Location.y + (bigatom_height - 1)

		if (top_x < 1 || top_y < 1 || top_x > world.maxx || top_y > world.maxy)
			return 0
		return 1
